# coding: utf-8
#
# author: codeskyblue
# created: 2016-05-16
#
# code reference from chromiumwebapps/chromium
#
# It is insteresting to know this way which can calculate not only fps, but also every frame use time.
# Is it really useful, I don't know, really don't know.
# so code just stop here.

import time
import subprocess


nanoseconds_per_second = 1e9

view = 'SurfaceView'

def init_frame_data():
    out = subprocess.check_output(['adb', 'shell', 'dumpsys', 'SurfaceFlinger', '--latency-clear', view])
    if out.strip() != '':
        raise RuntimeError("Not supported.")
    time.sleep(0.1)
    (refresh_period, timestamps) = frame_data()
    base_timestamp = 0
    base_index = 0
    for timestamp in timestamps:
        if timestamp != 0:
            base_timestamp = timestamp
            break
        base_index += 1

    if base_timestamp == 0:
        raise RuntimeError("Initial frame collect failed")
    return (refresh_period, base_timestamp, timestamps[base_index:])


def frame_data():
    out = subprocess.check_output(['adb', 'shell', 'dumpsys', 'SurfaceFlinger', '--latency', view])
    results = out.splitlines()
    refresh_period = long(results[0]) / nanoseconds_per_second
    timestamps = []
    for line in results[1:]:
        fields = line.split()
        if len(fields) != 3:
            continue
        (start, submitting, submitted) = map(int, fields)
        if submitting == 0:
            continue

        timestamp = submitting/nanoseconds_per_second
        timestamps.append(timestamp)
    return (refresh_period, timestamps)


def continue_collect_frame_data():
    refresh_period, base_timestamp, timestamps = init_frame_data()
    while True:
        refresh_period, tss = frame_data()
        last_index = 0
        if timestamps:
            recent_timestamp = timestamps[-2]
            last_index = tss.index(recent_timestamp)
        timestamps = timestamps[:-2] + tss[last_index:]
        time.sleep(1.5)
        ajusted_timestamps = []
        for seconds in timestamps[-4:]:
            seconds -= base_timestamp
            if seconds > 1e6: # too large, just ignore
                continue
            ajusted_timestamps.append(seconds)
        print ajusted_timestamps


if __name__ == '__main__':
    continue_collect_frame_data()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Blueprint, not finished yet.

>>> dev.packages()
[{'name': 'com.example.demo', 'version': 2}]
>>> dev.forward_list()
[{'local': 'tcp:8001', 'remote': 'tcp:8000'}]
>>> dev.properties()
{'ro.build.brand', 'MI2', ...}
>>> dev.install('demo.apk')
True
>>> dev.uninstall('com.example.demo', keep_data=True) # DONE
True
>>> dev.logcat() # TODO
>>> dev.pull('/data/local/tmp/_screen.png', './')
True
>>> dev.push('./demo.apk', '/data/local/tmp/demo.apk')
True
>>> dev.listdir('/data/local/tmp')
['_screen.png']
>>> dev.shell('ls', '-l', '/data/local/tmp/')
:output as string, replace \r\n to '\n'
>>> dev.start_activity('com.example.demo', '.Client')
None
>>> dev.stat('/data/local/tmp/_screen.png')
:posix.stat_result object
>>> dev.current_app()
com.example.demo
>>> dev.orientation()
: one of [1-4]
>>> dev.screenshot() # DONE
: PIL image object
>>> dev.keyevent('HOME')
None
>>> dev.open_minicap()
True
>>> dev.open_minitouch()
True
>>> dev.touch(100, 100)
None
>>> dev.swipe() # TODO
>>> dev.pinch() # only in minitouch
"""

from __future__ import absolute_import

from atx.adbkit.client import Client


if __name__ == '__main__':
    adb = Client()
    print adb.devices()
    print adb.version()
    dev = adb.device() #'10.250.210.165:57089')
    # print dev.keyevent('HOME')
    print dev.display
    # for pkg in dev.packages():
    #     print pkg
    # dev.screenshot('s.png', scale=1.0)
    print dev.is_locked()
    print dev.wake()
    dev.click(568, 1488)
    print dev.current_app()


# coding: utf-8
#
# dummy device is used for test

from __future__ import absolute_import

import os

from PIL import Image

from atx.device.mixin import DeviceMixin, hook_wrap
from atx.device import Display
from atx import consts


__dir__ = os.path.dirname(os.path.abspath(__file__))

class DummyDevice(DeviceMixin):
    def __init__(self, *args, **kwargs):
        DeviceMixin.__init__(self)
        self._display = Display(1280, 720)
        self._rotation = 1
        self.last_click = None
        self.serial = '1234'

    @hook_wrap(consts.EVENT_SCREENSHOT)
    def screenshot(self, filename=None):
        """ Take a screenshot """
        # screen size: 1280x720
        screen_path = os.path.join(__dir__, '../../tests/media/dummy_screen.png')
        screen = Image.open(screen_path)
        if filename:
            screen.save(filename)
        return screen

    @property
    def display(self):
        return self._display

    @property
    def rotation(self):
        return self._rotation

    def click(self, x, y):
        self.last_click = (x, y)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import threading
from functools import partial

from atx import logutils


log = logutils.getLogger(__name__) #base.getLogger('patch')


def run_once(f):
    ''' 
    Decorator: Make sure function only call once
    not thread safe

    @run_once
    def foo():
        print 'bar'
        return 1+2
    foo()
    foo() # 'bar' only print once
    '''
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.result = f(*args, **kwargs)
            wrapper.has_run = True
        return wrapper.result
    wrapper.has_run = False
    return wrapper

def attachmethod(target):
    '''
    Reference: https://blog.tonyseek.com/post/open-class-in-python/

    class Spam(object):
        pass

    @attach_method(Spam)
    def egg1(self, name):
        print((self, name))

    spam1 = Spam()
    # OpenClass 加入的方法 egg1 可用
    spam1.egg1("Test1")
    # 输出Test1
    '''
    if isinstance(target, type):
        def decorator(func):
            setattr(target, func.__name__, func)
    else:
        def decorator(func):
            setattr(target, func.__name__, partial(func, target))
    return decorator

def fuckit(fn):
    def decorator(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            args = list(args).extend([k+'='+v for k, v in kwargs.items()])
            print 'function(%s(%s)) panic(%s). fuckit' %(fn.__name__, ' ,'.join(args), e)
            return None
    return decorator

def go(fn):
    '''
    Decorator
    '''
    def decorator(*args, **kwargs):
        log.info('begin run func(%s) in background', fn.__name__)
        t = threading.Thread(target=fn, args=args, kwargs=kwargs)
        t.setDaemon(True)
        t.start()
        return t
    return decorator

# test code
if __name__ == '__main__':
    @go
    def say_hello(sleep=0.3, message='hello world'):
        time.sleep(sleep)
        print message
        return None
    t1 = say_hello(0.1)
    t2 = say_hello(0.5, 'this message should not showed')
    t1.join()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import

import atexit
import os
import re
import json
import collections
import tempfile
from StringIO import StringIO

from PIL import Image
from atx import logutils


logger = logutils.getLogger(__name__)

_DISPLAY_RE = re.compile(
    r'.*DisplayViewport{valid=true, .*orientation=(?P<orientation>\d+), .*deviceWidth=(?P<width>\d+), deviceHeight=(?P<height>\d+).*')
_PROP_PATTERN = re.compile(r'\[(?P<key>.*?)\]:\s*\[(?P<value>.*)\]')


class Device(object):
    Display = collections.namedtuple('Display', ['width', 'height', 'rotation'])
    Package = collections.namedtuple('Package', ['name', 'path'])
    __minicap = '/data/local/tmp/minicap'

    def __init__(self, client, serial):
        ''' TODO: client change to host, port '''
        self._client = client
        self._serial = serial
        self._screenshot_method = 'minicap'

    @property
    def serial(self):
        return self._serial
    
    def raw_cmd(self, *args, **kwargs):
        args = ['-s', self._serial] + list(args)
        return self._client.raw_cmd(*args)

    def run_cmd(self, *args):
        """
        Unix style output, already replace \r\n to \n
        """
        p = self.raw_cmd(*args)
        return p.communicate()[0].replace('\r\n', '\n')

    def shell(self, *args):
        """
        Run command `adb shell`
        """
        args = ['shell'] + list(args)
        return self.run_cmd(*args)

    def keyevent(self, key):
        ''' Call: adb shell input keyevent $key '''
        self.shell('input', 'keyevent', key)

    def remove(self, filename):
        """ 
        Remove file from device
        """
        output = self.shell('rm', filename)
        # any output means rm failed.
        return False if output else True

    def install(self, filename):
        """
        TOOD(ssx): Install apk into device, show progress

        Args:
            - filename(string): apk file path
        """
        return self.run_cmd('install', '-rt', filename)

    def uninstall(self, package_name, keep_data=False):
        """
        Uninstall package

        Args:
            - package_name(string): package name ex: com.example.demo
            - keep_data(bool): keep the data and cache directories
        """
        if keep_data:
            return self.run_cmd('uninstall', '-k', package_name)
        else:
            return self.run_cmd('uninstall', package_name)

    def pull(self, source_file, target_file=None):
        if target_file is None:
            raise RuntimeError('Not supported now')
        self.run_cmd('pull', source_file, target_file)

    @property
    def display(self):
        '''
        Return device width, height, rotation
        '''
        w, h = (0, 0)
        for line in self.shell('dumpsys', 'display').splitlines():
            m = _DISPLAY_RE.search(line, 0)
            if not m:
                continue
            w = int(m.group('width'))
            h = int(m.group('height'))
            o = int(m.group('orientation'))
            w, h = min(w, h), max(w, h)
            return self.Display(w, h, o)

        output = self.shell('LD_LIBRARY_PATH=/data/local/tmp', self.__minicap, '-i')
        try:
            data = json.loads(output)
            (w, h, o) = (data['width'], data['height'], data['rotation']/90)
            return self.Display(w, h, o)            
        except ValueError:
            pass

    def rotation(self):
        """
        Android rotation
        Return:
            - int [0-4]
        """
        return self.display.rotation
    
    def properties(self):
        '''
        Android Properties, extracted from `adb shell getprop`

        Returns:
            dict of props, for
            example:
                {'ro.bluetooth.dun': 'true'}
        '''
        props = {}
        for line in self.shell(['getprop']).splitlines():
            m = _PROP_PATTERN.match(line)
            if m:
                props[m.group('key')] = m.group('value')
        return props

    def packages(self):
        """
        Show all packages
        """
        pattern = re.compile(r'package:(/[^=]+\.apk)=([^\s]+)')
        packages = []
        for line in self.shell('pm', 'list', 'packages', '-f').splitlines():
            m = pattern.match(line)
            if not m:
                continue
            path, name = m.group(1), m.group(2)
            packages.append(self.Package(name, path))
        return packages

    def _adb_screencap(self, scale=1.0):
        """
        capture screen with adb shell screencap
        """
        remote_file = tempfile.mktemp(dir='/data/local/tmp/', prefix='screencap-', suffix='.png')
        local_file = tempfile.mktemp(prefix='atx-screencap-', suffix='.png')
        self.shell('screencap', '-p', remote_file)
        try:
            self.pull(remote_file, local_file)
            image = Image.open(local_file)
            image.load() # because Image is a lazy load function
            if scale is not None and scale != 1.0:
                image = image.resize([int(scale * s) for s in image.size], Image.BICUBIC)
            rotation = self.rotation()
            if rotation:
                method = getattr(Image, 'ROTATE_{}'.format(rotation*90))
                image = image.transpose(method)
            return image
        finally:
            self.remove(remote_file)
            os.unlink(local_file)

    def _adb_minicap(self, scale=1.0):
        """
        capture screen with minicap

        https://github.com/openstf/minicap
        """
        remote_file = tempfile.mktemp(dir='/data/local/tmp/', prefix='minicap-', suffix='.jpg')
        local_file = tempfile.mktemp(prefix='atx-minicap-', suffix='.jpg')
        (w, h, r) = self.display
        params = '{x}x{y}@{rx}x{ry}/{r}'.format(x=w, y=h, rx=int(w*scale), ry=int(h*scale), r=r*90)
        try:
            self.shell('LD_LIBRARY_PATH=/data/local/tmp', self.__minicap, '-s', '-P', params, '>', remote_file)
            self.pull(remote_file, local_file)
            with open(local_file, 'rb') as f:
                image = Image.open(StringIO(f.read()))
            return image
        finally:
            self.remove(remote_file)
            os.unlink(local_file)

    def screenshot(self, filename=None, scale=1.0, method=None):
        """
        Take device screenshot

        Args:
            - filename(string): optional, save int filename
            - scale(float): scale size
            - method(string): one of minicap,screencap

        Return:
            PIL.Image
        """
        image = None
        method = method or self._screenshot_method
        if method == 'minicap':
            try:
                image = self._adb_minicap(scale)
            except Exception as e:
                logger.warn("use minicap failed, fallback to screencap. error detail: %s", e)
                self._screenshot_method = 'screencap'
                return self.screenshot(filename=filename, scale=scale)
        elif method == 'screencap':
            image = self._adb_screencap(scale)
        else:
            raise RuntimeError("No such method(%s)" % method)

        if filename:
            image.save(filename)
        return image

    def click(self, x, y):
        '''
        same as adb -s ${SERIALNO} shell input tap x y
        FIXME(ssx): not tested on horizontal screen
        '''
        self.shell('input', 'tap', str(x), str(y))

    def forward(self, local_port, remote_port):
        '''
        adb port forward. return local_port
        TODO: not tested
        '''
        return self._client.forward(self.serial, local_port, remote_port)

    def is_locked(self):
        """
        Returns:
            - lock state(bool)
        Raises:
            RuntimeError
        """
        _lockScreenRE = re.compile('mShowingLockscreen=(true|false)')
        m = _lockScreenRE.search(self.shell('dumpsys', 'window', 'policy'))
        if m:
            return (m.group(1) == 'true')
        raise RuntimeError("Couldn't determine screen lock state")

    def is_screen_on(self):
        '''
        Checks if the screen is ON.
        Returns:
            True if the device screen is ON
        Raises:
            RuntimeError
        '''

        _screenOnRE = re.compile('mScreenOnFully=(true|false)')
        m = _screenOnRE.search(self.shell('dumpsys', 'window', 'policy'))
        if m:
            return (m.group(1) == 'true')
        raise RuntimeError("Couldn't determine screen ON state")

    def wake(self):
        """
        Wake up device if device locked
        """
        if not self.is_screen_on():
            self.keyevent('POWER')

    def is_keyboard_shown(self):
        dim = self.shell('dumpsys', 'input_method')
        if dim:
            # FIXME: API >= 15 ?
            return "mInputShown=true" in dim
        return False

    def current_app(self):
        """
        Return: (package_name, activity)
        Raises:
            RuntimeError
        """
        _focusedRE = re.compile('mFocusedApp=.*ActivityRecord{\w+ \w+ (?P<package>.*)/(?P<activity>.*) .*')
        m = _focusedRE.search(self.shell('dumpsys', 'window', 'windows'))
        if m:
            return m.group('package'), m.group('activity')
        raise RuntimeError("Couldn't get focused app")



#-*- encoding: utf-8 -*-

import bisect
import cv2
import os
import pickle
import Queue
import sys
import threading
import time
import traceback
import json

from collections import namedtuple

class BaseRecorder(object):

    monitor_period = 3 # seconds

    def __init__(self, device=None, workdir=".", realtime_analyze=False):
        self.device = None
        self.device_info = {}
        self.running = False
        self.setup_workdir(workdir)

        if device is not None:
            self.attach(device)
        self.realtime_analyze = realtime_analyze

        self.thread = None
        self.frames = []  # for backup
        self.last_frame_time = None
        self.case_draft = []  # for analyze draft
        self.input_queue = Queue.Queue()
        self.input_index = 0

        # find addons from base classes
        self.addons = {}
        for cls in self.__class__.__bases__:
            name = getattr(cls, '_%s__addon_name' % (cls.__name__,), None)
            if name is not None:
                gfun = getattr(self, 'get_%s' % (name,))
                sfun = getattr(self, 'save_%s' % (name,))
                lfun = getattr(self, 'load_%s' % (name,))
                self.addons[name] = (gfun, sfun, lfun)

    def setup_workdir(self, workdir):
        # setup direcoties
        self.workdir = workdir

        self.casedir = os.path.join(workdir, 'case')
        if not os.path.exists(self.casedir):
            os.makedirs(self.casedir)

        self.framedir = os.path.join(workdir, 'frames')
        if not os.path.exists(self.framedir):
            os.makedirs(self.framedir)

    def update_device_info(self):
        if self.device is None:
            return
        # TODO: define general device info
        w, h = self.device.display
        self.device_info = {"width":w, "height":h}

    def start(self):
        '''start running in background.'''
        self.update_device_info()
        self.get_device_status(0) # start addons.
        self.hook()
        self.thread = threading.Thread(target=self._run)
        self.thread.start()
        self.running = True

    def stop(self):
        self.unhook()
        self.running = False
        self.thread.join()

    def get_device_status(self, t):
        '''get device status at a given time t (within self.monitor_period)'''
        data = {}
        for name, (func, _, _) in self.addons.iteritems():
            data[name] = func(t)
        return data

    def _run(self):
        while True:
            try:
                time.sleep(0.1)
                frame = self.input_queue.get_nowait()
                self.handle_frame(frame)
            except Queue.Empty:
                if not self.running:
                    break
            except:
                traceback.print_exc()
                self.running = False

        # save meta info for backup & draft.
        if not self.realtime_analyze:
            self.analyze_all()
        self.save()
        print 'recorder stopped.'
        sys.exit()

    def input_event(self, event):
        '''should be called when user input events happens (from hook)'''
        if not self.running or self.device is None:
            return
        # print 'input_event', event.time
        status = self.get_device_status(event.time)
        self.input_queue.put((self.input_index, event, status))
        self.input_index += 1

    def handle_frame(self, frame):
        # print 'handle frame'
        idx, event, status = frame
        meta = {'index':idx}
        meta['event'] = self.serialize_event(event)
        if self.last_frame_time is None:
            meta['waittime'] = 0
        else:
            meta['waittime'] = event.time - self.last_frame_time
        self.last_frame_time = event.time

        # save frames.
        # print 'saving...'
        eventpath = os.path.join(self.framedir, '%d-event.pkl' % idx)
        pickle.dump(event, file(eventpath, 'w'))
        meta['status'] = {}
        for name, obj in status.iteritems():
            func = self.addons[name][1]
            data = func(obj, self.framedir, idx)
            meta['status'][name] = data

        # analyze
        if self.realtime_analyze:
            self.analyze_frame(idx, event, status)
        self.frames.append(meta)

    def serialize_event(self, event):
        return {}

    def analyze_frame(self, idx, event, status):
        '''analyze status and generate draft config'''
        # Example:
        #
        # d = {
        #     'action' : 'click',
        #     'args' : (100, 100),
        #     },
        # }
        # self.case_draft.append(d)
        pass

    def process_draft(self, d):
        '''generate code from draft'''
        raise NotImplementedError()

    def analyze_all(self):
        print 'total frames:', len(self.frames)
        print 'analying, please wait'
        for meta in self.frames:
            idx = meta['index']
            event = pickle.load(file(os.path.join(self.framedir, '%d-event.pkl' % idx)))
            status = meta['status']
            status = {}
            for name, data in meta['status'].iteritems():
                if data is None:
                    status[name] = None
                    continue
                func = self.addons[name][2]
                status[name] = func(self.framedir, data)
            self.analyze_frame(idx, event, status)
            print '\b.',
        print '\nDone'

    @classmethod
    def analyze_frames(cls, workdir):
        '''generate draft from recorded frames'''
        record = cls(None, workdir)
        obj = {}
        with open(os.path.join(workdir, 'frames', 'frames.json')) as f:
            obj = json.load(f)
        record.device_info = obj['device']
        record.frames = obj['frames']
        record.analyze_all()
        record.save()

    @classmethod
    def process_casefile(cls, workdir):
        '''generate code from case.json'''
        record = cls(None, workdir)
        obj = {}
        with open(os.path.join(workdir, 'frames', 'frames.json')) as f:
            obj = json.load(f)
        record.device_info = obj['device']
        record.frames = obj['frames']

        casedir = os.path.join(workdir, 'case')
        with open(os.path.join(casedir, 'case.json')) as f:
            record.case_draft = json.load(f)

        # remove old files
        for f in os.listdir(casedir):
            if f != 'case.json':
                os.remove(os.path.join(casedir, f))

        record.generate_script()

    def save(self):
        # save frames info, do not overwrite.
        filepath = os.path.join(self.framedir, 'frames.json')
        if not os.path.exists(filepath):
            obj = {
                'ctime' : time.ctime(),
                'device' : self.device_info,
                'frames' : self.frames,
            }
            with open(filepath, 'w') as f:
                json.dump(obj, f, indent=2)

        # save draft info
        filepath = os.path.join(self.framedir, 'draft.json')
        with open(filepath, 'w') as f:
            json.dump(self.case_draft, f, indent=2)
        # make a copy at casedir
        filepath = os.path.join(self.casedir, 'case.json')
        with open(filepath, 'w') as f:
            json.dump(self.case_draft, f, indent=2)

        # generate_script
        self.generate_script()

    def generate_script(self):
        # save draft pyscript
        print 'Generating script...'
        filepath = os.path.join(self.casedir, 'script.py')
        content = [
            '#-*- encoding: utf-8 -*-', '# Generated by recorder.',
            '',
            'import time',
            '',
            'def test(d):',
        ]
        for row in self.case_draft:
            script = self.process_draft(row).encode('utf-8', 'ignore')
            for line in script.split('\n'):
                content.append(' '*4 + line)
        content.extend([
            '',
            'if __name__ == "__main__":',
            '   import atx',
            '   d = atx.connect()',
            '   test(d)',
        ])
        with open(filepath, 'w') as f:
            f.write('\n'.join(content))
        print 'Save script to', filepath

    def attach(self, device):
        '''Attach to device, if current device is not None, should
        detach from it first. '''
        raise NotImplementedError()

    def detach(self):
        '''Detach from current device.'''
        raise NotImplementedError()

    def hook(self):
        '''Hook user input.'''
        raise NotImplementedError()

    def unhook(self):
        '''Unhook user input.'''
        raise NotImplementedError()

__CaptureRecord = namedtuple('__CaptureRecord', ('ctime', 'image'))

class CaptureRecord(__CaptureRecord):
    def __eq__(self, other):
        return self[0] == other[0]
    def __ne__(self, other):
        return not self == other
    def __gt__(self, other):
        return self[0] > other[0]
    def __ge__(self, other):
        return not self<other
    def __lt__(self, other):
        return self[0] < other[0]
    def __le__(self, other):
        return not other<self

class ScreenAddon(object):
    __addon_name = 'screen'

    __capture_cache = []
    __capture_interval = 0.1
    __capture_lock = None
    __capture_thread = None

    def get_screen(self, t):
        if self.__capture_thread is None:
            self.__start()
        with self.__capture_lock:
            idx = bisect.bisect(self.__capture_cache, (t, None))
            if idx != 0:
                return self.__capture_cache[idx-1][1]

    def save_screen(self, screen, dirpath, idx):
        if screen is None:
            return
        filename = '%d.png' % idx
        filepath = os.path.join(dirpath, filename)
        cv2.imwrite(filepath, screen)
        return filename

    def load_screen(self, dirpath, filename):
        filepath = os.path.join(dirpath, filename)
        try:
            return cv2.imread(filepath)
        except:
            return

    def __start(self):
        print 'start', self.__addon_name
        if self.__capture_lock is None:
            self.__capture_lock = threading.Lock()
        if self.__capture_thread is not None:
            self.__capture_thread._Thread_stop() # using __stop private method, not good
        self.__capture_thread = t = threading.Thread(target=self.__capture)
        t.setDaemon(True)
        t.start()

    def __capture(self):
        capture_maxnum = int(self.monitor_period/self.__capture_interval)
        while True:
            self.__capture_lock.acquire()
            try:
                time.sleep(self.__capture_interval)
                if not self.running or self.device is None:
                    continue
                # tic = time.time()
                img = self.device.screenshot_cv2()
                # print '--capturing.. cost', time.time() - tic
                self.__capture_cache.append(CaptureRecord(time.time(), img))
                self.__capture_cache = self.__capture_cache[-capture_maxnum:]
            finally:
                self.__capture_lock.release()

class UixmlAddon(object):
    __addon_name = 'uixml'

    __uidump_cache = []
    __uidump_interval = 0.1
    __uidump_lock = None
    __uidump_thread = None

    def get_uixml(self, t):
        if self.__uidump_thread is None:
            self.__start()
        with self.__uidump_lock:
            idx = bisect.bisect(self.__uidump_cache, (t, ''))
            if idx != 0:
                return self.__uidump_cache[idx-1][1]

    def save_uixml(self, uixml, dirpath, idx):
        if uixml is None:
            return
        filename = '%d-uidump.xml' % idx
        filepath = os.path.join(dirpath, filename)
        with open(filepath, 'w') as f:
            f.write(uixml)
        return filename

    def load_uixml(self, dirpath, filename):
        filepath = os.path.join(dirpath, filename)
        try:
            return open(filepath).read()
        except IOError:
            return u''

    def __start(self):
        print 'start', self.__addon_name
        if self.__uidump_lock is None:
            self.__uidump_lock = threading.Lock()
        if self.__uidump_thread is not None:
            self.__uidump_thread._Thread_stop() # using __stop private method, not good
        self.__uidump_thread = t = threading.Thread(target=self.__dump)
        t.setDaemon(True)
        t.start()

    def __dump(self):
        uidump_maxnum = int(self.monitor_period/self.__uidump_interval)
        while True:
            self.__uidump_lock.acquire()
            try:
                time.sleep(self.__uidump_interval)
                if not self.running or self.device is None:
                    continue
                # tic = time.time()
                xmldata = self.device.dumpui()
                xmldata = xmldata.encode('utf-8')
                # print 'dumping ui.. cost', time.time() - tic
                self.__uidump_cache.append((time.time(), xmldata))
                self.__uidump_cache = self.__uidump_cache[-uidump_maxnum:]
            finally:
                self.__uidump_lock.release()


if __name__ == '__main__':

    class TestRecorder(BaseRecorder, ScreenAddon, UixmlAddon):
        def attach(self, device):
            self.device = device
        def detach(self): pass
        def hook(self): pass
        def unhook(self): pass

    class DummyDevice(object):
        def screenshot_cv2(self):
            return None

        def dumpui(self):
            return u'uixml'

    class DummyEvent(object):
        def __init__(self):
            self.time = time.time()

    r = TestRecorder(DummyDevice(), 'testcase')
    r.start()

    count = 10
    while count > 0:
        try:
            time.sleep(1)
            e = DummyEvent()
            r.input_event(e)
        except:
            break
        count -= 1

    r.stop()
